"""The Python `cf` package is an Earth Science data analysis library
that is built on a complete implementation of the CF data model.

The `cf` package implements the CF data model for its internal data
structures and so is able to process any CF-compliant dataset. It is
not strict about CF-compliance, however, so that partially conformant
datasets may be ingested from existing datasets and written to new
datasets. This is so that datasets which are partially conformant may
nonetheless be modified in memory.

The `cf` package can:

* read field constructs and domain constructs from netCDF, CDL, PP and
  UM datasets,

* create new field and domain constructs in memory,

* inspect field and domain constructs,

* test whether two constructs are the same,

* modify field and domain construct metadata and data,

* create subspaces of field and domain constructs,

* write and append field constructs to netCDF datasets on disk,

* incorporate, and create, metadata stored in external files,

* read, write, and create data that have been compressed by convention
  (i.e. ragged or gathered arrays), whilst presenting a view of the
  data in its uncompressed form,

* read, write, and create coordinates defined by geometry cells,

* read netCDF and CDL datasets containing hierarchical groups,

* combine field constructs arithmetically,

* manipulate field construct data by arithmetical and trigonometrical
  operations,

* perform statistical collapses on field constructs,

* perform histogram, percentile and binning operations on field
  constructs,

* regrid field constructs with (multi-)linear, nearest neighbour,
  first- and second-order conservative and higher order patch recovery
  methods,

* apply convolution filters to field constructs,

* calculate derivatives of field constructs,

* create field constructs to create derived quantities (such as
  vorticity).

All of the above use LAMA functionality, which allows multiple fields
larger than the available memory to exist and be manipulated. (Note:
work is underway to replace this functionality with a dask
implementation.)


**Visualisation**

Powerful, flexible, and very simple to produce visualisations of field
constructs uses the `cfplot` package
(http://ajheaps.github.io/cf-plot), that is automatically installed
along with with `cf`.

See the :ref:`cf-python home page <cf-python-home>` for documentation,
installation and source code.

"""

__Conventions__ = "CF-1.10"
__date__ = "2023-01-31"
__version__ = "3.14.0b0"

_requires = (
    "numpy",
    "netCDF4",
    "cftime",
    "cfunits",
    "cfdm",
    "psutil",
    "dask",
    "packaging",
)

x = ", ".join(_requires)
_error0 = f"cf v{ __version__} requires the modules {x}. "

try:
    import cfdm
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

__cf_version__ = cfdm.core.__cf_version__

from packaging.version import Version
import importlib.util
import platform

_found_ESMF = bool(importlib.util.find_spec("ESMF"))

try:
    import netCDF4
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import numpy
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import cftime
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import cfunits
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import psutil
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import dask
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

try:
    import packaging
except ImportError as error1:
    raise ImportError(_error0 + str(error1))

# Check the version of packaging
_minimum_vn = "20.0"
if Version(packaging.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad packaging version: cf requires packaging>={_minimum_vn}. "
        f"Got {packaging.__version__} at {packaging.__file__}"
    )

# Check the version of psutil
_minimum_vn = "0.6.0"
if Version(psutil.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad psutil version: cf requires psutil>={_minimum_vn}. "
        f"Got {psutil.__version__} at {psutil.__file__}"
    )

# Check the version of netCDF4
_minimum_vn = "1.5.4"
if Version(netCDF4.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad netCDF4 version: cf requires netCDF4>={_minimum_vn}. "
        f"Got {netCDF4.__version__} at {netCDF4.__file__}"
    )

# Check the version of cftime
_minimum_vn = "1.6.0"
if Version(cftime.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad cftime version: cf requires cftime>={_minimum_vn}. "
        f"Got {cftime.__version__} at {cftime.__file__}"
    )

# Check the version of numpy
_minimum_vn = "1.22"
if Version(numpy.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad numpy version: cf requires numpy>={_minimum_vn}. "
        f"Got {numpy.__version__} at {numpy.__file__}"
    )

# Check the version of cfunits
_minimum_vn = "3.3.4"
if Version(cfunits.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad cfunits version: cf requires cfunits>={_minimum_vn}. "
        f"Got {cfunits.__version__} at {cfunits.__file__}"
    )

# Check the version of cfdm
_minimum_vn = "1.10.0.0"
_maximum_vn = "1.10.1.0"
_cfdm_version = Version(cfdm.__version__)
if not Version(_minimum_vn) <= _cfdm_version < Version(_maximum_vn):
    raise RuntimeError(
        f"Bad cfdm version: cf requires {_minimum_vn}<=cfdm<{_maximum_vn}. "
        f"Got {_cfdm_version} at {cfdm.__file__}"
    )

# Check the version of dask
_minimum_vn = "2022.12.1"
if Version(dask.__version__) < Version(_minimum_vn):
    raise RuntimeError(
        f"Bad dask version: cf requires dask>={_minimum_vn}. "
        f"Got {dask.__version__} at {dask.__file__}"
    )

# Check the version of Python
_minimum_vn = "3.7.0"
if Version(platform.python_version()) < Version(_minimum_vn):
    raise ValueError(
        f"Bad python version: cf requires python version {_minimum_vn} "
        f"or later. Got {platform.python_version()}"
    )

from .constructs import Constructs

from .mixin import Coordinate

from .count import Count
from .index import Index
from .interpolationparameter import InterpolationParameter
from .list import List
from .nodecountproperties import NodeCountProperties
from .partnodecountproperties import PartNodeCountProperties
from .interiorring import InteriorRing
from .tiepointindex import TiePointIndex

from .bounds import Bounds
from .domain import Domain
from .datum import Datum
from .coordinateconversion import CoordinateConversion

from .cfdatetime import dt, dt_vector
from .flags import Flags
from .timeduration import TimeDuration, Y, M, D, h, m, s
from .units import Units

from .constructlist import ConstructList
from .fieldlist import FieldList
from .domainlist import DomainList

from .dimensioncoordinate import DimensionCoordinate
from .auxiliarycoordinate import AuxiliaryCoordinate
from .coordinatereference import CoordinateReference
from .cellmethod import CellMethod
from .cellmeasure import CellMeasure
from .domainancillary import DomainAncillary
from .domainaxis import DomainAxis
from .fieldancillary import FieldAncillary
from .field import Field
from .data import Data
from .data.array import (
    CFANetCDFArray,
    FullArray,
    GatheredArray,
    NetCDFArray,
    RaggedContiguousArray,
    RaggedIndexedArray,
    RaggedIndexedContiguousArray,
    SubsampledArray,
)

from .data.fragment import (
    MissingFragmentArray,
    NetCDFFragmentArray,
    UMFragmentArray,
)

from .aggregate import aggregate
from .query import (
    Query,
    lt,
    le,
    gt,
    ge,
    eq,
    ne,
    contain,
    contains,
    wi,
    wo,
    set,
    year,
    month,
    day,
    hour,
    minute,
    second,
    dtlt,
    dtle,
    dtgt,
    dtge,
    dteq,
    dtne,
    cellsize,
    cellge,
    cellgt,
    cellle,
    celllt,
    cellwi,
    cellwo,
    djf,
    mam,
    jja,
    son,
    seasons,
)

from .constants import *  # noqa: F403
from .functions import *  # noqa: F403
from .maths import curl_xy, div_xy, relative_vorticity, histogram
from .examplefield import example_field, example_fields, example_domain

from .cfimplementation import CFImplementation, implementation

from .read_write import read, write

from .regrid import RegridOperator


# Set up basic logging for the full project with a root logger
import logging
import sys

# Configure the root logger which all module loggers inherit from:
logging.basicConfig(
    stream=sys.stdout,
    style="{",  # default is old style ('%') string formatting
    format="{message}",  # no module names or datetimes etc. for basic case
    level=logging.WARNING,  # default but change level via log_level()
)

# And create custom level inbetween 'INFO' & 'DEBUG', to understand value see:
# https://docs.python.org/3.8/howto/logging.html#logging-levels
logging.DETAIL = 15  # set value as an attribute as done for built-in levels
logging.addLevelName(logging.DETAIL, "DETAIL")


def detail(self, message, *args, **kwargs):
    if self.isEnabledFor(logging.DETAIL):
        self._log(logging.DETAIL, message, args, **kwargs)


logging.Logger.detail = detail



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class ConstructsTest(unittest.TestCase):
    """Unit test for the Constructs class."""

    f = cf.example_field(1)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cf.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cf.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cf.log_level('DISABLE')

    def test_Constructs__repr__(self):
        """Test all means of Construct inspection."""
        f = self.f

        repr(f.constructs)

    def test_Constructs_filter_by_naxes(self):
        """Test the `filter_by_naxes` Constructs method."""
        c = self.f.constructs

        self.assertEqual(len(c.filter_by_naxes()), 12)
        self.assertEqual(len(c.filter_by_naxes(1)), 7)
        self.assertEqual(len(c.filter_by_naxes(cf.ge(2))), 5)
        self.assertEqual(len(c.filter_by_naxes(1, cf.ge(2))), 12)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print("")
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class RegridTest(unittest.TestCase):
    filename1 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file1.nc"
    )
    filename2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file2.nc"
    )
    filename3 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file3.nc"
    )
    filename4 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file4.nc"
    )
    filename5 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file3.nc"
    )
    filename6 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file2.nc"
    )
    filename7 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file5.nc"
    )
    filename8 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file6.nc"
    )
    filename9 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file7.nc"
    )
    filename10 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file8.nc"
    )

    chunk_sizes = (300, 10000, 100000)[::-1]

    @unittest.skipUnless(cf._found_ESMF, "Requires ESMF package.")
    def test_Field_regrids(self):
        self.assertFalse(cf.regrid_logging())
        with cf.atol(1e-12):
            for chunksize in self.chunk_sizes:
                with cf.chunksize(chunksize):
                    f1 = cf.read(self.filename1)[0]
                    f2 = cf.read(self.filename2)[0]
                    f3 = cf.read(self.filename3)[0]
                    f4 = cf.read(self.filename4)[0]
                    f5 = cf.read(self.filename5)[0]

                    r = f1.regrids(f2, "conservative")
                    self.assertTrue(
                        f3.equals(r),
                        f"destination=global Field, CHUNKSIZE={chunksize}",
                    )

                    dst = {"longitude": f2.dim("X"), "latitude": f2.dim("Y")}
                    r = f1.regrids(dst, "conservative", dst_cyclic=True)
                    self.assertTrue(
                        f3.equals(r),
                        f"destination=global dict, CHUNKSIZE={chunksize}",
                    )

                    r = f1.regrids(dst, method="conservative", dst_cyclic=True)
                    self.assertTrue(
                        f3.equals(r),
                        f"destination=global dict, CHUNKSIZE={chunksize}",
                    )

                    # Regrid global to regional roated pole
                    r = f1.regrids(f5, method="linear")
                    self.assertTrue(
                        f4.equals(r, verbose=3),
                        f"destination=regional Field, CHUNKSIZE={chunksize}",
                    )

        f6 = cf.read(self.filename6)[0]
        with self.assertRaises(Exception):
            f1.regridc(f6, axes="T", method="linear")

    @unittest.skipUnless(cf._found_ESMF, "Requires ESMF package.")
    def test_Field_regridc(self):
        self.assertFalse(cf.regrid_logging())
        with cf.atol(1e-11):
            for chunksize in self.chunk_sizes:
                self.assertFalse(cf.regrid_logging())
                with cf.chunksize(chunksize):
                    f1 = cf.read(self.filename7)[0]
                    f2 = cf.read(self.filename8)[0]
                    f3 = cf.read(self.filename9)[0]
                    self.assertTrue(
                        f3.equals(f1.regridc(f2, axes="T", method="linear")),
                        f"destination=time series, CHUNKSIZE={chunksize}",
                    )
                    f4 = cf.read(self.filename1)[0]
                    f5 = cf.read(self.filename2)[0]
                    f6 = cf.read(self.filename10)[0]
                    self.assertTrue(
                        f6.equals(
                            f4.regridc(
                                f5, axes=("X", "Y"), method="conservative"
                            )
                        ),
                        f"destination=global Field, CHUNKSIZE={chunksize}",
                    )
                    self.assertTrue(
                        f6.equals(
                            f4.regridc(
                                f5, axes=("X", "Y"), method="conservative"
                            )
                        ),
                        f"destination=global Field, CHUNKSIZE={chunksize}",
                    )
                    dst = {"X": f5.dim("X"), "Y": f5.dim("Y")}
                    self.assertTrue(
                        f6.equals(
                            f4.regridc(
                                dst, axes=("X", "Y"), method="conservative"
                            )
                        ),
                        f"destination=global dict, CHUNKSIZE={chunksize}",
                    )
                    self.assertTrue(
                        f6.equals(
                            f4.regridc(
                                dst, axes=("X", "Y"), method="conservative"
                            )
                        ),
                        f"destination=global dict, CHUNKSIZE={chunksize}",
                    )

    @unittest.skipUnless(cf._found_ESMF, "Requires ESMF package.")
    def test_Field_regrids_operator(self):
        self.assertFalse(cf.regrid_logging())

        with cf.atol(1e-12):
            f1 = cf.read(self.filename1)[0]
            f2 = cf.read(self.filename2)[0]
            f3 = cf.read(self.filename3)[0]
            f4 = cf.read(self.filename4)[0]
            f5 = cf.read(self.filename5)[0]

            op = f1.regrids(f2, "conservative", return_operator=True)
            r = f1.regrids(op)
            self.assertTrue(f3.equals(r))

            # Repeat
            r = f1.regrids(op)
            self.assertTrue(f3.equals(r))

            dst = {"longitude": f2.dim("X"), "latitude": f2.dim("Y")}
            op = f1.regrids(
                dst, "conservative", dst_cyclic=True, return_operator=True
            )
            r = f1.regrids(op)
            self.assertTrue(f3.equals(r))

            op = f1.regrids(
                dst,
                method="conservative",
                dst_cyclic=True,
                return_operator=True,
            )
            r = f1.regrids(op)
            self.assertTrue(f3.equals(r))

            # Regrid global to regional rotated pole
            op = f1.regrids(f5, method="linear", return_operator=True)
            r = f1.regrids(op)
            self.assertTrue(f4.equals(r))

        # Raise exception when the source grid does not match that of
        # the regrid operator
        op = f1.regrids(f2, "conservative", return_operator=True)
        with self.assertRaises(ValueError):
            f2.regrids(op)

    @unittest.skipUnless(cf._found_ESMF, "Requires ESMF package.")
    def test_Field_regridc_operator(self):
        self.assertFalse(cf.regrid_logging())

        with cf.atol(1e-12):
            f1 = cf.read(self.filename7)[0]
            f2 = cf.read(self.filename8)[0]
            f3 = cf.read(self.filename9)[0]
            f4 = cf.read(self.filename1)[0]
            f5 = cf.read(self.filename2)[0]
            f6 = cf.read(self.filename10)[0]

            op = f1.regridc(
                f2, axes="T", method="linear", return_operator=True
            )
            self.assertTrue(f3.equals(f1.regridc(op)))

            op = f4.regridc(
                f5,
                axes=("X", "Y"),
                method="conservative",
                return_operator=True,
            )
            self.assertTrue(f6.equals(f4.regridc(op)))

            op = f4.regridc(
                f5,
                axes=("X", "Y"),
                method="conservative",
                return_operator=True,
            )
            self.assertTrue(f6.equals(f4.regridc(op)))

            dst = {
                "X": f5.dimension_coordinate("X"),
                "Y": f5.dimension_coordinate("Y"),
            }
            op = f4.regridc(
                dst,
                axes=("X", "Y"),
                method="conservative",
                return_operator=True,
            )

            self.assertTrue(f6.equals(f4.regridc(op)))
            self.assertTrue(f6.equals(f4.regridc(op)))

        # Raise exception when the source grid does not match that of
        # the regrid operator
        op = f1.regridc(f2, axes="T", method="linear", return_operator=True)
        with self.assertRaises(ValueError):
            f2.regrids(op)

    @unittest.skipUnless(cf._found_ESMF, "Requires ESMF package.")
    def test_Field_regrid_size1_dimensions(self):
        # Check that non-regridded size 1 axes are handled OK
        self.assertFalse(cf.regrid_logging())

        f = cf.example_field(0)
        shape = f.shape

        g = f.regrids(f, method="linear")
        self.assertEqual(g.shape, (shape))
        g = f.regridc(f, method="linear", axes="X")
        self.assertEqual(g.shape, (shape))

        f.insert_dimension("T", position=0, inplace=True)
        shape = f.shape
        g = f.regrids(f, method="linear")
        self.assertEqual(g.shape, shape)
        g = f.regridc(f, method="linear", axes="X")
        self.assertEqual(g.shape, shape)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print("")
    unittest.main(verbosity=2)



import datetime
import faulthandler
import unittest

import cftime
import dask.array as da
import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class DataUtilsTest(unittest.TestCase):
    def test_Data_Utils__da_ma_allclose(self):
        """TODO."""
        # Create a range of inputs to test against.
        # Note that 'a' and 'a2' should be treated as 'allclose' for this
        # method, the same result as np.ma.allclose would give because all
        # of the *unmasked* elements are 'allclose', whereas in our
        # Data.equals method that builds on this method, we go even further
        # and insist on the mask being identical as well as the data
        # (separately, i.e. unmasked) all being 'allclose', so inside our
        # cf.Data objects 'a' and 'a2' would instead *not* be considered equal.
        a_np = np.ma.array([1.0, 2.0, 3.0], mask=[1, 0, 0])
        a = da.from_array(a_np)
        a2 = da.from_array(np.ma.array([10.0, 2.0, 3.0], mask=[1, 0, 0]))
        b_np = np.ma.array([1.0, 2.0, 3.0], mask=[0, 1, 0])
        b = da.from_array(b_np)
        c_np = np.ma.array([1.0, 2.0, 100.0], mask=[1, 0, 0])
        c = da.from_array(c_np)
        d = da.from_array(np.array([1.0, 2.0, 3.0]))
        e = a + 5e-04  # outside of tolerance to set, namely rtol=1e-05
        f = a + 5e-06  # within set tolerance to be specified, as above

        # Test the function with these inputs as both numpy and dask arrays...
        allclose = cf.data.dask_utils._da_ma_allclose

        self.assertTrue(allclose(a, a).compute())
        self.assertTrue(allclose(a2, a).compute())
        self.assertTrue(allclose(b, a).compute())

        # ...including testing the 'masked_equal' parameter
        self.assertFalse(allclose(b, a, masked_equal=False).compute())

        self.assertFalse(allclose(c, a).compute())
        self.assertTrue(allclose(d, a).compute())
        self.assertFalse(allclose(e, a).compute())

        self.assertTrue(allclose(f, a, rtol=1e-05).compute())

        # Test when array inputs have different chunk sizes
        a_chunked = da.from_array(a_np, chunks=(1, 2))
        self.assertTrue(
            allclose(da.from_array(b_np, chunks=(3,)), a_chunked).compute()
        )
        self.assertFalse(
            allclose(
                da.from_array(b_np, chunks=(3,)), a_chunked, masked_equal=False
            ).compute()
        )
        self.assertFalse(
            allclose(da.from_array(c_np, chunks=(3,)), a_chunked).compute()
        )

        # Test the 'rtol' and 'atol' parameters:
        self.assertFalse(allclose(e, a, rtol=1e-06).compute())
        b1 = e / 10000
        b2 = a / 10000
        self.assertTrue(allclose(b1, b2, atol=1e-05).compute())

    def test_Data_Utils__is_numeric_dtype(self):
        """TODO."""
        _is_numeric_dtype = cf.data.utils._is_numeric_dtype
        for a in [
            np.array([0, 1, 2]),
            np.array([False, True, True]),
            np.ma.array([10.0, 2.0, 3.0], mask=[1, 0, 0]),
            np.array(10),
        ]:
            self.assertTrue(_is_numeric_dtype(a))

        for b in [
            np.array(["a", "b", "c"], dtype="S1"),
            np.empty(1, dtype=object),
        ]:
            self.assertFalse(_is_numeric_dtype(b))

    def test_Data_Utils_convert_to_datetime(self):
        """TODO."""
        a = cftime.DatetimeGregorian(2000, 12, 3, 12)
        for x in (2.5, [2.5]):
            d = da.from_array(x)
            e = cf.data.utils.convert_to_datetime(
                d, cf.Units("days since 2000-12-01")
            )
            self.assertEqual(e.compute(), a)

        a = [
            cftime.DatetimeGregorian(2000, 12, 1),
            cftime.DatetimeGregorian(2000, 12, 2),
            cftime.DatetimeGregorian(2000, 12, 3),
        ]
        for x in ([0, 1, 2], [[0, 1, 2]]):
            d = da.from_array([0, 1, 2], chunks=2)
            e = cf.data.utils.convert_to_datetime(
                d, cf.Units("days since 2000-12-01")
            )
            self.assertTrue((e.compute() == a).all())

    def test_Data_Utils_convert_to_reftime(self):
        """TODO."""
        a = cftime.DatetimeGregorian(2000, 12, 3, 12)
        d = da.from_array(np.array(a, dtype=object))

        e, u = cf.data.utils.convert_to_reftime(d)
        self.assertEqual(e.compute(), 0.5)
        self.assertEqual(u, cf.Units("days since 2000-12-03", "standard"))

        units = cf.Units("days since 2000-12-01")
        e, u = cf.data.utils.convert_to_reftime(d, units=units)
        self.assertEqual(e.compute(), 2.5)
        self.assertEqual(u, units)

        a = "2000-12-03T12:00"
        d = da.from_array(np.array(a, dtype=str))

        e, u = cf.data.utils.convert_to_reftime(d)
        self.assertEqual(e.compute(), 0.5)
        self.assertEqual(u, cf.Units("days since 2000-12-03", "standard"))

        units = cf.Units("days since 2000-12-01")
        e, u = cf.data.utils.convert_to_reftime(d, units=units)
        self.assertEqual(e.compute(), 2.5)
        self.assertEqual(u, units)

        a = [
            [
                cftime.DatetimeGregorian(2000, 12, 1),
                cftime.DatetimeGregorian(2000, 12, 2),
                cftime.DatetimeGregorian(2000, 12, 3),
            ]
        ]
        d = da.from_array(np.ma.array(a, mask=[[1, 0, 0]]), chunks=2)

        e, u = cf.data.utils.convert_to_reftime(d)
        self.assertTrue((e.compute() == [-99, 0, 1]).all())
        self.assertEqual(u, cf.Units("days since 2000-12-02", "standard"))

        units = cf.Units("days since 2000-12-03")
        e, u = cf.data.utils.convert_to_reftime(d, units=units)
        self.assertTrue((e.compute() == [-99, -1, 0]).all())
        self.assertEqual(u, units)

        d = cf.Data(
            ["2004-02-29", "2004-02-30", "2004-03-01"], calendar="360_day"
        )
        self.assertEqual(d.Units, cf.Units("days since 2004-02-29", "360_day"))
        self.assertTrue((d.array == [0, 1, 2]).all())

        d = cf.Data(["2004-02-29", "2004-03-01"], dt=True)
        self.assertEqual(d.Units, cf.Units("days since 2004-02-29"))
        self.assertTrue((d.array == [0, 1]).all())

    def test_Data_Utils_unique_calendars(self):
        """TODO."""
        a = [
            [
                cftime.DatetimeGregorian(2000, 12, 1),
                cftime.DatetimeGregorian(2000, 12, 2),
                cftime.DatetimeGregorian(2000, 12, 3),
            ]
        ]
        d = da.from_array(np.ma.array(a, mask=[[1, 0, 0]]), chunks=2)
        c = cf.data.utils.unique_calendars(d)
        self.assertIsInstance(c, set)
        self.assertEqual(c, set(["standard"]))

        a = cftime.DatetimeGregorian(2000, 12, 1)
        d = da.from_array(np.array(a, dtype=object))
        c = cf.data.utils.unique_calendars(d)
        self.assertEqual(c, set(["standard"]))

        d[()] = np.ma.masked
        c = cf.data.utils.unique_calendars(d)
        self.assertEqual(c, set())

        a = [
            cftime.DatetimeGregorian(2000, 12, 1),
            cftime.DatetimeAllLeap(2000, 12, 2),
            cftime.DatetimeGregorian(2000, 12, 3),
        ]
        d = da.from_array(np.ma.array(a, mask=[1, 0, 0]), chunks=2)
        c = cf.data.utils.unique_calendars(d)
        self.assertEqual(c, set(["all_leap", "standard"]))

    def test_Data_Utils_first_non_missing_value(self):
        """TODO."""
        for method in ("index", "mask"):
            # Scalar data
            d = da.from_array(0)
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), 0
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            d[()] = np.ma.masked
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), None
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            # 1-d data
            d = da.arange(8)
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), 0
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            d[0] = np.ma.masked
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), 1
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            # 2-d data
            d = da.arange(8).reshape(2, 4)
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), 0
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            d[0] = np.ma.masked
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), 4
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

            d[...] = np.ma.masked
            self.assertEqual(
                cf.data.utils.first_non_missing_value(d, method=method), None
            )
            self.assertEqual(
                cf.data.utils.first_non_missing_value(
                    d, cached=99, method=method
                ),
                99,
            )

        # Bad method
        with self.assertRaises(ValueError):
            cf.data.utils.first_non_missing_value(d, method="bad")

    def test_Data_Utils_conform_units(self):
        for x in (1, [1, 2], "foo", np.array([[1]])):
            self.assertEqual(cf.data.utils.conform_units(x, cf.Units("m")), x)

        d = cf.Data([1000, 2000], "m")
        e = cf.data.utils.conform_units(d, cf.Units("m"))
        self.assertIs(e, d)
        e = cf.data.utils.conform_units(d, cf.Units("km"))
        self.assertTrue(e.equals(cf.Data([1, 2], "km"), ignore_data_type=True))

        with self.assertRaises(ValueError):
            cf.data.utils.conform_units(d, cf.Units("s"))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import ctypes as CT
import os

import numpy
import numpy.ctypeslib

from . import umfile

_len_real_hdr = 19
_len_int_hdr = 45


class File_type(CT.Structure):
    _fields_ = [
        ("fmt", CT.c_int),
        ("byte_ordering", CT.c_int),
        ("word_size", CT.c_int),
    ]


def _get_ctypes_array(dtype, size=None):
    """Get ctypes corresponding to a numpy array of a given type.

    The size should not be necessary unless the storage for the array is
    allocated in the C code.

    """
    kwargs = {
        "dtype": dtype,
        "ndim": 1,
        "flags": ("C_CONTIGUOUS", "WRITEABLE"),
    }
    if size:
        kwargs["shape"] = (size,)

    return numpy.ctypeslib.ndpointer(**kwargs)


def _gen_rec_class(int_type, float_type):
    class Rec(CT.Structure):
        """ctypes object corresponding to the `Rec` object in the C
        code."""

        _fields_ = [
            ("int_hdr", _get_ctypes_array(int_type, _len_int_hdr)),
            ("real_hdr", _get_ctypes_array(float_type, _len_real_hdr)),
            ("header_offset", CT.c_size_t),
            ("data_offset", CT.c_size_t),
            ("disk_length", CT.c_size_t),
            ("_internp", CT.c_void_p),
        ]

    return Rec


Rec32 = _gen_rec_class(numpy.int32, numpy.float32)
Rec64 = _gen_rec_class(numpy.int64, numpy.float64)


def _gen_var_class(rec_class):
    class Var(CT.Structure):
        """ctypes object corresponding to the `Var` object in the C
        code."""

        _fields_ = [
            ("recs", CT.POINTER(CT.POINTER(rec_class))),
            ("nz", CT.c_int),
            ("nt", CT.c_int),
            ("supervar_index", CT.c_int),
            ("_internp", CT.c_void_p),
        ]

    return Var


Var32 = _gen_var_class(Rec32)
Var64 = _gen_var_class(Rec64)


def _gen_file_class(var_class):
    class File(CT.Structure):
        """ctypes object corresponding to the `File` object in the C
        code."""

        _fields_ = [
            ("fd", CT.c_int),
            ("file_type", File_type),
            ("nvars", CT.c_int),
            ("vars", CT.POINTER(CT.POINTER(var_class))),
            ("_internp", CT.c_void_p),
        ]

    return File


File32 = _gen_file_class(Var32)
File64 = _gen_file_class(Var64)


class Enum:
    def __init__(self, *names):
        self.names = names

    def as_name(self, val):
        if isinstance(val, str):
            return val
        else:
            return self.names[val]

    def as_index(self, val):
        if isinstance(val, int):
            return val

        return self.names.index(val)


enum_file_format = Enum("PP", "FF")
enum_byte_ordering = Enum("little_endian", "big_endian")
enum_data_type = Enum("integer", "real")


class CInterface:
    """Interface to the C shared library functions."""

    def __init__(self, lib_name="umfile.so"):
        """**Initialisation**

        :Parameters:

            lib_name: `str`
                The name of the C library binary.

        """
        lib_dir = os.path.join(os.path.dirname(__file__) or ".", "c-lib")
        lib_path = os.path.join(lib_dir, lib_name)
        self.lib = CT.CDLL(lib_path)

    def _is_null_pointer(self, ptr):
        """True if the pointer is a null pointer.

        :Returns:

            `bool`

        """
        try:
            ptr.contents
            return False
        except ValueError:
            return True

    def detect_file_type(self, fd):
        """Auto-detect file type.

        :Parameters:

            fd: `int`
                The file descriptor of the open file.

        :Returns:

            `File_type`
                A `File_type` ctypes object that can be passed to
                `file_parse`, or raises an exception if file type cannot
                be detected.

        """
        file_type = File_type()
        rv = self.lib.detect_file_type(fd, CT.pointer(file_type))
        if rv != 0:
            raise umfile.UMFileException("File type could not be detected")

        return file_type

    def file_type_obj_to_dict(self, file_type):
        """Converts a `File_type` object returned by `detect_file_type` into a
        dictionary that includes meaningful string values in place of the
        integers that derive from the C enum statments, specifically:
        'fmt': 'PP' or 'FF' 'byte_ordering': 'little_endian' or
        'big_endian' and also 'word_size': 4 or 8

        :Returns:

            `dict`

        """
        fmt = enum_file_format.as_name(file_type.fmt)
        byte_ordering = enum_byte_ordering.as_name(file_type.byte_ordering)
        word_size = file_type.word_size
        return {
            "fmt": fmt,
            "byte_ordering": byte_ordering,
            "word_size": word_size,
        }

    def create_file_type(self, fmt, byte_ordering, word_size):
        """Creates a `File_type` object for passing to `file_parse`.

        :Parameters:

            fmt: `str`
                'PP' or 'FF'

            byte_ordering: `str`
                'little_endian' or 'big_endian'

            word_size: `str`
                4 or 8

        :Returns:

            `File_type`
                A `File_type` object (ctypes structure containing integer
                values) that can be passed to `file_parse`.

        """
        return File_type(
            fmt=enum_file_format.as_index(fmt),
            byte_ordering=enum_byte_ordering.as_index(byte_ordering),
            word_size=word_size,
        )

    def set_word_size(self, val):
        """Sets the word size used to interpret returned pointers from
        subsequent calls, in particular the pointers to PP headers
        embedded in the tree of objects returned by `file_parse` and the
        data array that is populated by `read_record_data`.

        :Parameters:

             val: `int` or `File_type`
                 Either just the word_size value to use or a `File_type`
                 object from which it is to be extracted.

        """
        if isinstance(val, File_type):
            word_size = val.word_size
        else:
            word_size = val

        if word_size == 4:
            self.file_class = File32
            self.file_data_int_type = numpy.int32
            self.file_data_real_type = numpy.float32
            self._int_ptr = CT.POINTER(CT.c_int32)
            self._real_ptr = CT.POINTER(CT.c_float)
        elif word_size == 8:
            self.file_class = File64
            self.file_data_int_type = numpy.int64
            self.file_data_real_type = numpy.float64
            self._int_ptr = CT.POINTER(CT.c_int64)
            self._real_ptr = CT.POINTER(CT.c_double)
        else:
            raise ValueError(f"Word size must be 4 or 8 (not {word_size!r})")

    def _get_ctypes_int_array(self, size=None):
        """Get ctypes corresponding to the `numpy` integer array."""
        return _get_ctypes_array(self.file_data_int_type, size)

    def _get_ctypes_real_array(self, size=None):
        """Get ctypes corresponding to the `numpy` real array."""
        return _get_ctypes_array(self.file_data_real_type, size)

    def _get_empty_real_array(self, size):
        """Get empty `numpy` real array according to word size
        previously set with `set_word_size`."""
        return numpy.empty(size, dtype=self.file_data_real_type)

    def _get_empty_int_array(self, size):
        """Get empty `numpy` integer array according to word size
        previously set with `set_word_size`."""
        return numpy.empty(size, dtype=self.file_data_int_type)

    def parse_file(self, fh, file_type):
        """Given an open file handle, work out information from the
        file.

        :Parameters:

            fh: `int`
                Low-level file handle.

            file_type: `File_type`
                `File_type` object as returned by `detect_file_type` or
                `create_file_type`.

        :Returns:

            `dict`
                The information from the file. Currently the only key
                actually implemented is ``'vars'``, containing a list of
                variables, as that is all that the caller requires.

        """
        func = self.lib.file_parse
        file_p_type = CT.POINTER(self.file_class)
        func.restype = file_p_type

        file_p = func(fh, file_type)
        if self._is_null_pointer(file_p):
            raise umfile.UMFileException("File parsing failed")

        file = file_p.contents
        c_vars = file.vars[: file.nvars]
        rv = {"vars": list(map(self.c_var_to_py_var, c_vars))}

        # Now that we have copied all the data into python objects for
        # the caller, free any memory allocated in the C code before
        # returning
        free_func = self.lib.file_free
        free_func._fields_ = file_p_type
        free_func(file_p)

        return rv

    def c_var_to_py_var(self, c_var_p):
        """Create a `umfile.Var` object from a ctypes object
        corresponding to 'Var*' in the C code.

        :Returns:

            `umfile.Var`

        """
        c_var = c_var_p.contents
        nz = c_var.nz
        nt = c_var.nt
        svi = c_var.supervar_index
        if svi < 0:
            svi = None

        c_recs = c_var.recs
        recs = [
            self.c_rec_to_py_rec(c_recs[recid]) for recid in range(nz * nt)
        ]

        return umfile.Var(recs, nz, nt, svi)

    def c_rec_to_py_rec(self, c_rec_p):
        """Create a `umfile.Rec` object from a ctypes object
        corresponding to 'Rec*' in the C code.

        :Returns:

            `umfile.Rec`

        """
        c_rec = c_rec_p.contents

        # ============================================================
        # Previous code - causing memory leaks per
        # https://github.com/numpy/numpy/issues/6511
        #
        #     # numpy.copy used here so we can go back and free the memory
        #     # allocated by C without affecting the python object
        #     int_hdr = numpy.copy(numpy.ctypeslib.as_array(c_rec.int_hdr))
        #     real_hdr = numpy.copy(numpy.ctypeslib.as_array(c_rec.real_hdr))
        # ============================================================

        #
        # -------------------------------
        # Workaround: instead cast to a pointer, obtain the values knowing
        # the length of the header arrays, and copy into an appropriate
        # numpy array
        #
        ptr = CT.cast(c_rec.int_hdr, self._int_ptr)
        int_hdr = numpy.array(
            ptr[:_len_int_hdr], dtype=self.file_data_int_type
        )

        ptr = CT.cast(c_rec.real_hdr, self._real_ptr)
        real_hdr = numpy.array(
            ptr[:_len_real_hdr], dtype=self.file_data_real_type
        )
        # ============================================================

        header_offset = c_rec.header_offset
        data_offset = c_rec.data_offset
        disk_length = c_rec.disk_length
        return umfile.Rec(
            int_hdr, real_hdr, header_offset, data_offset, disk_length
        )

    def get_type_and_num_words(self, int_hdr):
        """From the integer header, work out data type and number of
        words to read (`read_record_data` requires this).

        :Returns:

           `str`, `int`
               The datatype ('integer' or 'real') and the number of words.

        """
        word_size = int_hdr.itemsize
        self.lib.get_type_and_num_words.argtypes = [
            CT.c_int,
            self._get_ctypes_int_array(),
            CT.POINTER(CT.c_int),
            CT.POINTER(CT.c_size_t),
        ]
        data_type = CT.c_int()
        num_words = CT.c_size_t()
        rv = self.lib.get_type_and_num_words(
            word_size, int_hdr, CT.pointer(data_type), CT.pointer(num_words)
        )
        if rv != 0:
            raise umfile.UMFileException(
                "Error determining data type and size from integer header"
            )

        return enum_data_type.as_name(data_type.value), num_words.value

    def get_extra_data_offset_and_length(
        self, int_hdr, data_offset, disk_length
    ):
        """From the integer header, gets offset and length of extra
        data.

        :Returns:

            `int`, `int`
                The offset and length, both in units of BYTES.

        """
        word_size = int_hdr.itemsize
        func = self.lib.get_extra_data_offset_and_length
        func.argtypes = [
            CT.c_int,
            self._get_ctypes_int_array(),
            CT.c_size_t,
            CT.c_size_t,
            CT.POINTER(CT.c_size_t),
            CT.POINTER(CT.c_size_t),
        ]
        extra_data_offset = CT.c_size_t()
        extra_data_length = CT.c_size_t()
        rv = func(
            word_size,
            int_hdr,
            data_offset,
            disk_length,
            CT.pointer(extra_data_offset),
            CT.pointer(extra_data_length),
        )
        if rv != 0:
            raise umfile.UMFileException(
                "Error determining extra data length from integer header"
            )

        return extra_data_offset.value, extra_data_length.value

    def read_header(self, fd, header_offset, byte_ordering, word_size):
        """Reads the header from open file.

        :Returns:

            `numpy.ndarray`, `numpy.ndarray`
                The integer and real parts of the header.

        """
        self.lib.read_header.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            self._get_ctypes_int_array(),
            self._get_ctypes_real_array(),
        ]

        int_hdr = self._get_empty_int_array(_len_int_hdr)
        real_hdr = self._get_empty_real_array(_len_real_hdr)
        rv = self.lib.read_header(
            fd,
            header_offset,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            int_hdr,
            real_hdr,
        )
        if rv != 0:
            raise umfile.UMFileException("Error reading header data")

        return int_hdr, real_hdr

    def read_extra_data(
        self,
        fd,
        extra_data_offset,
        extra_data_length,
        byte_ordering,
        word_size,
    ):
        """Reads record data from open file.

        inputs:
           fd - integer low-level file descriptor
           extra_data_offset - offset in bytes
           extra_disk_length - disk length of extra data in bytes
           byte_ordering - 'little_endian' or 'big_endian'
           word_size - 4 or 8

        returns: extra data as string

        """
        extra_data = b"\0" * extra_data_length

        self.lib.read_extra_data.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            CT.c_char_p,
        ]

        rv = self.lib.read_extra_data(
            fd,
            extra_data_offset,
            extra_data_length,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            extra_data,
        )
        if rv != 0:
            raise umfile.UMFileException("Error reading extra data")

        return extra_data

    def read_record_data(
        self,
        fd,
        data_offset,
        disk_length,
        byte_ordering,
        word_size,
        int_hdr,
        real_hdr,
        data_type,
        nwords,
    ):
        """Reads record data from open file.

        inputs:
           fd - integer low-level file descriptor
           data_offset - offset in words
           disk_length - disk length of data record in words
           byte_ordering - 'little_endian' or 'big_endian'
           word_size - 4 or 8
           int_hdr - integer PP headers (numpy array)
           real_hdr - real PP headers (numpy array)
           data_type - 'integer' or 'real'
           nwords - number of words to read
              type and nwords should have been returned by
              get_type_and_num_words()

        """
        if data_type == "integer":
            data = self._get_empty_int_array(nwords)
            ctypes_data = self._get_ctypes_int_array()
        elif data_type == "real":
            data = self._get_empty_real_array(nwords)
            ctypes_data = self._get_ctypes_real_array()
        else:
            raise ValueError("data_type must be 'integer' or 'real'")

        self.lib.read_record_data.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            self._get_ctypes_int_array(),
            self._get_ctypes_real_array(),
            CT.c_size_t,
            ctypes_data,
        ]

        rv = self.lib.read_record_data(
            fd,
            data_offset,
            disk_length,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            int_hdr,
            real_hdr,
            nwords,
            data,
        )

        if rv != 0:
            raise umfile.UMFileException("Error reading record data")

        return data


if __name__ == "__main__":
    import sys

    c = CInterface()
    fd = os.open(sys.argv[1], os.O_RDONLY)
    file_type = c.detect_file_type(fd)
    print(c.file_type_obj_to_dict(file_type))
    c.set_word_size(file_type)
    info = c.parse_file(fd, file_type)

    for var in info["vars"]:
        print("nz = %s, nt = %s" % (var.nz, var.nt))
        for rec in var.recs:
            print(rec.hdr_offset)
            print("data offset", rec.data_offset)
            print("disk length", rec.disk_length)
            print("int hdr", rec.int_hdr)
            print("real hdr", rec.real_hdr)
            data_type, nwords = c.get_type_and_num_words(rec.int_hdr)
            print("data_type = %s nwords = %s" % (data_type, nwords))

            data = c.read_record_data(
                fd,
                rec.data_offset,
                rec.disk_length,
                file_type.byte_ordering,
                file_type.word_size,
                rec.int_hdr,
                rec.real_hdr,
                data_type,
                nwords,
            )
            print(
                "data (%s values): %s ... %s"
                % (len(data), data[:10], data[-10:])
            )
            (
                extra_data_offset,
                extra_data_length,
            ) = c.get_extra_data_offset_and_length(
                rec.int_hdr, rec.data_offset, rec.disk_length
            )
            print("extra data offset: %s" % extra_data_offset)
            print("extra data length: %s" % extra_data_length)
            extra_data = c.read_extra_data(
                fd,
                extra_data_offset,
                extra_data_length,
                file_type.byte_ordering,
                file_type.word_size,
            )
            print("extra data (%s bytes) read" % (len(extra_data)))

    print(
        c.read_header(
            fd,
            info["vars"][0].recs[0].hdr_offset,
            file_type.byte_ordering,
            file_type.word_size,
        )
    )



"""Functions used during `Data` object collapses."""
import inspect
from functools import partial, reduce
from operator import mul

import numpy as np
from cfdm.core import DocstringRewriteMeta
from dask.array import chunk
from dask.array.core import _concatenate2
from dask.array.reductions import divide, numel, reduction
from dask.core import flatten
from dask.utils import deepmap

from ..docstring import _docstring_substitution_definitions


class Collapse(metaclass=DocstringRewriteMeta):
    """Container for functions that collapse dask arrays.

    .. versionadded:: TODODASKVER

    """

    def __docstring_substitutions__(self):
        """Define docstring substitutions that apply to this class and
        all of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionadded:: TODODASKVER

        .. seealso:: `_docstring_substitutions`

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Return the package depth for "package" docstring
        substitutions.

        See `_docstring_package_depth` for details.

        .. versionadded:: TODODASKVER

        """
        return 0

    @classmethod
    def max(cls, a, axis=None, keepdims=False, mtol=None, split_every=None):
        """Return maximum values of an array.

        Calculates the maximum value of an array or the maximum values
        along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = a.dtype
        return reduction(
            a,
            cf_max_chunk,
            partial(cf_max_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_max_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )

    @classmethod
    def max_abs(
        cls, a, axis=None, keepdims=False, mtol=None, split_every=None
    ):
        """Return maximum absolute values of an array.

        Calculates the maximum absolute value of an array or the
        maximum absolute values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        return cls.max(
            abs(a),
            axis=axis,
            keepdims=keepdims,
            mtol=mtol,
            split_every=split_every,
        )

    @classmethod
    def mean(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return mean values of an array.

        Calculates the mean value of an array or the mean values along
        axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = "f8"
        return reduction(
            a,
            cf_mean_chunk,
            partial(cf_mean_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_mean_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def mean_abs(
        cls,
        a,
        weights=None,
        axis=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return mean absolute values of an array.

        Calculates the mean absolute value of an array or the mean
        absolute values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        return cls.mean(
            abs(a),
            weights=weights,
            axis=axis,
            keepdims=keepdims,
            mtol=mtol,
            split_every=split_every,
        )

    @classmethod
    def mid_range(
        cls,
        a,
        axis=None,
        dtype=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return mid-range values of an array.

        Calculates the mid-range value of an array or the mid-range
        values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a, allowed="fi")
        dtype = "f8"
        return reduction(
            a,
            cf_range_chunk,
            partial(cf_mid_range_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_range_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )

    @classmethod
    def min(cls, a, axis=None, keepdims=False, mtol=None, split_every=None):
        """Return minimum values of an array.

        Calculates the minimum value of an array or the minimum values
        along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = a.dtype
        return reduction(
            a,
            cf_min_chunk,
            partial(cf_min_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_min_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )

    @classmethod
    def min_abs(
        cls, a, axis=None, keepdims=False, mtol=None, split_every=None
    ):
        """Return minimum absolute values of an array.

        Calculates the minimum absolute value of an array or the
        minimum absolute values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        return cls.min(
            abs(a),
            axis=axis,
            keepdims=keepdims,
            mtol=mtol,
            split_every=split_every,
        )

    @classmethod
    def range(cls, a, axis=None, keepdims=False, mtol=None, split_every=None):
        """Return range values of an array.

        Calculates the range value of an array or the range values
        along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a, allowed="fi")
        dtype = a.dtype
        return reduction(
            a,
            cf_range_chunk,
            partial(cf_range_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_range_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )

    @classmethod
    def rms(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return root mean square (RMS) values of an array.

        Calculates the RMS value of an array or the RMS values along
        axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = "f8"
        return reduction(
            a,
            cf_rms_chunk,
            partial(cf_rms_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_mean_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def sample_size(
        cls, a, axis=None, keepdims=False, mtol=None, split_every=None
    ):
        """Return sample size values of an array.

        Calculates the sample size value of an array or the sample
        size values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = "i8"
        return reduction(
            a,
            cf_sample_size_chunk,
            partial(cf_sample_size_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_sample_size_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )

    @classmethod
    def sum(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return sum values of an array.

        Calculates the sum value of an array or the sum values along
        axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = double_precision_dtype(a)
        if weights is not None:
            dtype = np.result_type(double_precision_dtype(weights), dtype)

        return reduction(
            a,
            cf_sum_chunk,
            partial(cf_sum_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_sum_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def sum_of_weights(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return sum of weights values for an array.

        Calculates the sum of weights value for an array or the sum of
        weights values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = double_precision_dtype(weights, default="i8")
        return reduction(
            a,
            cf_sum_of_weights_chunk,
            partial(cf_sum_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_sum_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def sum_of_weights2(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        split_every=None,
    ):
        """Return sum of squares of weights values for an array.

        Calculates the sum of squares of weights value for an array or
        the sum of squares of weights values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = double_precision_dtype(weights, default="i8")
        return reduction(
            a,
            partial(cf_sum_of_weights_chunk, square=True),
            partial(cf_sum_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_sum_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def var(
        cls,
        a,
        axis=None,
        weights=None,
        keepdims=False,
        mtol=None,
        ddof=None,
        split_every=None,
    ):
        """Return variances of an array.

        Calculates the variance value of an array or the variance
        values along axes.

        See
        https://ncas-cms.github.io/cf-python/analysis.html#collapse-methods
        for mathematical definitions.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{Collapse weights: data_like or `None`, optional}}

            {{collapse axes: (sequence of) `int`, optional}}

            {{collapse keepdims: `bool`, optional}}

            {{mtol: number, optional}}

            {{ddof: number}}

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The collapsed array.

        """
        check_input_dtype(a)
        dtype = "f8"
        return reduction(
            a,
            partial(cf_var_chunk, ddof=ddof),
            partial(cf_var_agg, mtol=mtol, original_shape=a.shape),
            axis=axis,
            keepdims=keepdims,
            dtype=dtype,
            split_every=split_every,
            combine=cf_var_combine,
            concatenate=False,
            meta=np.array((), dtype=dtype),
            weights=weights,
        )

    @classmethod
    def unique(cls, a, split_every=None):
        """Return unique elements of the data.

        .. versionadded:: TODODASKVER

        :Parameters:

            a: `dask.array.Array`
                The array to be collapsed.

            {{split_every: `int` or `dict`, optional}}

        :Returns:

            `dask.array.Array`
                The unique values in a 1-d array.

        """
        check_input_dtype(a, "fibUS")

        # Flatten the array so that it has the same number of
        # dimensions as the result (i.e. 1). This ensures that the
        # combination of `keepdims=True, output_size=np.nan` will
        # result in a correct output chunk size `np.nan`. See
        # `dask.array.reduction` for details.
        a = a.flatten()

        dtype = a.dtype
        return reduction(
            a,
            cf_unique_chunk,
            cf_unique_agg,
            keepdims=True,
            output_size=np.nan,
            dtype=dtype,
            split_every=split_every,
            concatenate=False,
            meta=np.array((), dtype=dtype),
        )


def check_input_dtype(a, allowed="fib"):
    """Check that data has a data type allowed by a collapse method.

    The collapse method is assumed to be defined by the calling
    function.

    :Parameters:

        a: `dask.array.Array`
            The data.

        allowed: `str`, optional
            The data type kinds allowed by the collapse
            method. Defaults to ``'fib'``, meaning that only float,
            integer and Boolean data types are allowed.

    :Returns:

        `None`

    """
    if a.dtype.kind not in allowed:
        method = inspect.currentframe().f_back.f_code.co_name
        raise TypeError(f"Can't calculate {method} of data with {a.dtype!r}")


def double_precision_dtype(a, default=None, bool_type="i"):
    """Returns the corresponding double precision data type of an array.

    :Parameters:

        a: `dask.array.Array` or `None`
            The data. If `None` then the value of *default* is
            returned*.

        default: `str`, optional
            If *a* is `None`, then return this data type.

        bool_type: `str`, optional
            The corresponding double data type kind for Boolean
            data. Defaults to ``'i'``, meaning ``'i8'`` is
            returned. Set to ``'f'` to return ``'f8'`` instead.

    :Returns:

        `str`
            The double precision type.

    **Examples**

    >>> for d in (int, 'int32', float, 'float32', bool):
    ...     print(double_precision_dtype(np.array(1, dtype=d)))
    ...
    i8
    i8
    f8
    f8
    i8

    >>> double_precision_dtype(np.array(1, dtype=bool), bool_type='f')
    'f8'
    >>> double_precision_dtype(None, default="i8")
    'i8'

    """
    if a is None:
        return default

    kind = a.dtype.kind
    if kind == "b":
        return bool_type + "8"

    if kind in "fi":
        return kind + "8"

    raise TypeError(f"Can't collapse data with {a.dtype!r}")


def mask_small_sample_size(x, N, axis, mtol, original_shape):
    """Mask elements where the sample size is below a threshold.

    .. versionadded:: TODODASKVER

    :Parameters:

        x: `numpy.ndarray`
            The collapsed data.

        N: `numpy.ndarray`
            The sample sizes of the collapsed values.

        axis: sequence of `int`
            The axes being collapsed.

        mtol: number
            The sample size threshold below which collapsed values are
            set to missing data. It is defined as a fraction (between
            0 and 1 inclusive) of the contributing input data values.

            The default of *mtol* is 1, meaning that a missing datum
            in the output array occurs whenever all of its
            contributing input array elements are missing data.

            For other values, a missing datum in the output array
            occurs whenever more than ``100*mtol%`` of its
            contributing input array elements are missing data.

            Note that for non-zero values of *mtol*, different
            collapsed elements may have different sample sizes,
            depending on the distribution of missing data in the input
            data.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

    :Returns:

        `numpy.ndarray`
            Array *x* masked where *N* is sufficiently small. Note
            that the input *x* might be modified in-place with the
            contents of the output.

    """
    if not x.ndim:
        # Make sure that we have a numpy array (e.g. as opposed to a
        # numpy.float64)
        x = np.asanyarray(x)

    if mtol < 1:
        # Nmax = total number of elements, including missing values
        Nmax = reduce(mul, [original_shape[i] for i in axis], 1)
        x = np.ma.masked_where(N < (1 - mtol) * Nmax, x, copy=False)

    return x


def sum_weights_chunk(x, weights=None, square=False, N=None, **kwargs):
    """Sum the weights.

    .. versionadded:: TODODASKVER

    :Parameters:

        x: `numpy.ndarray`
            The data.

        weights: `numpy.ndarray`, optional
            The weights associated with values of the data. Must have
            the same shape as *x*. By default *weights* is `None`,
            meaning that all non-missing elements of the data have a
            weight of 1 and all missing elements have a weight of
            0. If given as an array then those weights that are
            missing data, or that correspond to the missing elements
            of the data, are assigned a weight of 0.

        square: `bool`, optional
            If True calculate the sum of the squares of the weights.

        N: `numpy.ndarray`, optional
            The sample size. If provided as an array and there are no
            weights, then the *N* is returned instead of calculating
            the sum (of the squares) of weights. Ignored of *weights*
            is not `None`.

    :Returns:

        `numpy.ndarray`
            The sum of the weights, with data type "i8" or "f8".

    """
    if weights is None:
        # All weights are 1, so the sum of the weights and the sum of
        # the squares of the weights are both equal to the sample
        # size.
        if N is None:
            N = cf_sample_size_chunk(x, **kwargs)["N"]

        return N

    dtype = double_precision_dtype(weights)
    if square:
        weights = np.multiply(weights, weights, dtype=dtype)

    if np.ma.is_masked(x):
        weights = np.ma.masked_where(x.mask, weights)

    return chunk.sum(weights, dtype=dtype, **kwargs)


def combine_arrays(
    pairs, key, func, axis, dtype=None, computing_meta=False, **kwargs
):
    """Worker function for Combine callables.

    Select arrays by dictionary key from a nested list of
    dictionaries, concatenate the resulting nested list of arrays
    along the specified axes, and then apply a function to the result
    along those same axes.

    See `dask.array.reductions.mean_combine` for an example.

    .. versionadded:: TODODASKVER

    :Returns:

        `numpy.ndarray`

    """
    x = deepmap(lambda pair: pair[key], pairs) if not computing_meta else pairs

    if dtype:
        kwargs["dtype"] = dtype

    x = _concatenate2(x, axes=axis)
    return func(x, axis=axis, **kwargs)


def sum_arrays(pairs, key, axis, dtype, computing_meta=False, **kwargs):
    """Alias of `combine_arrays` with ``func=chunk.sum``.

    .. versionadded:: TODODASKVER

    """
    return combine_arrays(
        pairs, key, chunk.sum, axis, dtype, computing_meta, **kwargs
    )


def max_arrays(pairs, key, axis, dtype, computing_meta=False, **kwargs):
    """Alias of `combine_arrays` with ``func=chunk.max``.

    .. versionadded:: TODODASKVER

    """
    return combine_arrays(
        pairs, key, chunk.max, axis, dtype, computing_meta, **kwargs
    )


def min_arrays(pairs, key, axis, dtype, computing_meta=False, **kwargs):
    """Alias of `combine_arrays` with ``func=chunk.min``.

    .. versionadded:: TODODASKVER

    """
    return combine_arrays(
        pairs, key, chunk.min, axis, dtype, computing_meta, **kwargs
    )


def sum_sample_sizes(pairs, axis, computing_meta=False, **kwargs):
    """Alias of `combine_arrays` with ``key="N", func=chunk.sum,
    dtype="i8"``.

    .. versionadded:: TODODASKVER

    """
    return combine_arrays(
        pairs,
        "N",
        chunk.sum,
        axis,
        dtype="i8",
        computing_meta=computing_meta,
        **kwargs,
    )


# --------------------------------------------------------------------
# mean
# --------------------------------------------------------------------
def cf_mean_chunk(x, weights=None, dtype="f8", computing_meta=False, **kwargs):
    """Chunk calculations for the mean.

     This function is passed to `dask.array.reduction` as its *chunk*
     parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * V1: The sum of ``weights`` (equal to ``N`` if weights
                  are not set).
            * sum: The weighted sum of ``x``.
            * weighted: True if weights have been set.

    """
    if computing_meta:
        return x

    # N, sum
    d = cf_sum_chunk(x, weights, dtype=dtype, **kwargs)

    d["V1"] = sum_weights_chunk(x, weights, N=d["N"], **kwargs)
    d["weighted"] = weights is not None

    return d


def cf_mean_combine(
    pairs, axis=None, dtype="f8", computing_meta=False, **kwargs
):
    """Combination calculations for the mean.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_mean_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    weighted = next(flatten(pairs))["weighted"]
    d = {"weighted": weighted}

    d["sum"] = sum_arrays(pairs, "sum", axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d["sum"]

    d["N"] = sum_sample_sizes(pairs, axis, **kwargs)
    if weighted:
        d["V1"] = sum_arrays(pairs, "V1", axis, dtype, **kwargs)
    else:
        d["V1"] = d["N"]

    return d


def cf_mean_agg(
    pairs,
    axis=None,
    dtype="f8",
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the mean.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_mean_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = divide(d["sum"], d["V1"], dtype=dtype)
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# maximum
# --------------------------------------------------------------------
def cf_max_chunk(x, dtype=None, computing_meta=False, **kwargs):
    """Chunk calculations for the maximum.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * max: The maximum of `x``.

    """
    if computing_meta:
        return x

    return {
        "max": chunk.max(x, **kwargs),
        "N": cf_sample_size_chunk(x, **kwargs)["N"],
    }


def cf_max_combine(pairs, axis=None, computing_meta=False, **kwargs):
    """Combination calculations for the maximum.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_max_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    mx = max_arrays(pairs, "max", axis, None, computing_meta, **kwargs)
    if computing_meta:
        return mx

    return {"max": mx, "N": sum_sample_sizes(pairs, axis, **kwargs)}


def cf_max_agg(
    pairs,
    axis=None,
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the maximum.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_max_combine(pairs, axis, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = d["max"]
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# mid-range
# --------------------------------------------------------------------
def cf_mid_range_agg(
    pairs,
    axis=None,
    dtype="f8",
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the mid-range.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_range_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    # Calculate the mid-range
    x = divide(d["max"] + d["min"], 2.0, dtype=dtype)
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# minimum
# --------------------------------------------------------------------
def cf_min_chunk(x, dtype=None, computing_meta=False, **kwargs):
    """Chunk calculations for the minimum.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * min: The minimum of ``x``.

    """
    if computing_meta:
        return x

    return {
        "min": chunk.min(x, **kwargs),
        "N": cf_sample_size_chunk(x, **kwargs)["N"],
    }


def cf_min_combine(pairs, axis=None, computing_meta=False, **kwargs):
    """Combination calculations for the minimum.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_min_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    mn = min_arrays(pairs, "min", axis, None, computing_meta, **kwargs)
    if computing_meta:
        return mn

    return {"min": mn, "N": sum_sample_sizes(pairs, axis, **kwargs)}


def cf_min_agg(
    pairs,
    axis=None,
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the minimum.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_min_combine(pairs, axis, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = d["min"]
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# range
# --------------------------------------------------------------------
def cf_range_chunk(x, dtype=None, computing_meta=False, **kwargs):
    """Chunk calculations for the range.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * min: The minimum of ``x``.
            * max: The maximum of ``x`.

    """
    if computing_meta:
        return x

    # N, max
    d = cf_max_chunk(x, **kwargs)

    d["min"] = chunk.min(x, **kwargs)
    return d


def cf_range_combine(
    pairs, axis=None, dtype=None, computing_meta=False, **kwargs
):
    """Combination calculations for the range.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_range_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    mx = max_arrays(pairs, "max", axis, None, computing_meta, **kwargs)
    if computing_meta:
        return mx

    mn = min_arrays(pairs, "min", axis, None, **kwargs)

    return {"max": mx, "min": mn, "N": sum_sample_sizes(pairs, axis, **kwargs)}


def cf_range_agg(
    pairs,
    axis=None,
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the range.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_range_combine(pairs, axis, computing_meta, **kwargs)
    if computing_meta:
        return d

    # Calculate the range
    x = d["max"] - d["min"]
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# root mean square
# --------------------------------------------------------------------
def cf_rms_chunk(x, weights=None, dtype="f8", computing_meta=False, **kwargs):
    """Chunk calculations for the root mean square (RMS).

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * sum: The weighted sum of ``x**2``.

    """
    if computing_meta:
        return x

    return cf_mean_chunk(
        np.multiply(x, x, dtype=dtype), weights=weights, dtype=dtype, **kwargs
    )


def cf_rms_agg(
    pairs,
    axis=None,
    dtype="f8",
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the root mean square (RMS).

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_mean_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = np.sqrt(d["sum"] / d["V1"], dtype=dtype)
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# sample size
# --------------------------------------------------------------------
def cf_sample_size_chunk(x, dtype="i8", computing_meta=False, **kwargs):
    """Chunk calculations for the sample size.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.

    """
    if computing_meta:
        return x

    if np.ma.isMA(x):
        N = chunk.sum(np.ones_like(x, dtype=dtype), **kwargs)
    else:
        if dtype:
            kwargs["dtype"] = dtype

        N = numel(x, **kwargs)

    return {"N": N}


def cf_sample_size_combine(
    pairs, axis=None, dtype="i8", computing_meta=False, **kwargs
):
    """Combination calculations for the sample size.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_sample_size_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    x = sum_arrays(pairs, "N", axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return x

    return {"N": x}


def cf_sample_size_agg(
    pairs,
    axis=None,
    computing_meta=False,
    dtype="i8",
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the sample size.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_sample_size_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = d["N"]
    x = mask_small_sample_size(x, x, axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# sum
# --------------------------------------------------------------------
def cf_sum_chunk(x, weights=None, dtype="f8", computing_meta=False, **kwargs):
    """Chunk calculations for the sum.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * sum: The weighted sum of ``x``

    """
    if computing_meta:
        return x

    if weights is not None:
        x = np.multiply(x, weights, dtype=dtype)

    d = cf_sample_size_chunk(x, **kwargs)
    d["sum"] = chunk.sum(x, dtype=dtype, **kwargs)
    return d


def cf_sum_combine(
    pairs, axis=None, dtype="f8", computing_meta=False, **kwargs
):
    """Combination calculations for the sum.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_sum_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    x = sum_arrays(pairs, "sum", axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return x

    return {"sum": x, "N": sum_sample_sizes(pairs, axis, **kwargs)}


def cf_sum_agg(
    pairs,
    axis=None,
    dtype="f8",
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the sum.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_sum_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    x = d["sum"]
    x = mask_small_sample_size(x, d["N"], axis, mtol, original_shape)
    return x


# --------------------------------------------------------------------
# sum of weights
# --------------------------------------------------------------------
def cf_sum_of_weights_chunk(
    x, weights=None, dtype="f8", computing_meta=False, square=False, **kwargs
):
    """Chunk calculations for the sum of the weights.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    :Parameters:

        square: `bool`, optional
            If True then calculate the sum of the squares of the
            weights.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * sum: The sum of ``weights``, or the sum of
                   ``weights**2`` if *square* is True.

    """
    if computing_meta:
        return x

    # N
    d = cf_sample_size_chunk(x, **kwargs)

    d["sum"] = sum_weights_chunk(
        x, weights=weights, square=square, N=d["N"], **kwargs
    )

    return d


# --------------------------------------------------------------------
# unique
# --------------------------------------------------------------------
def cf_unique_chunk(x, dtype=None, computing_meta=False, **kwargs):
    """Chunk calculations for the unique values.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

    See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * unique: The unique values.

    """
    if computing_meta:
        return x

    return {"unique": np.unique(x)}


def cf_unique_agg(pairs, axis=None, computing_meta=False, **kwargs):
    """Aggregation calculations for the unique values.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    It is assumed that the arrays are one-dimensional.

    .. versionadded:: TODODASKVER

    :Parameters:

    See `dask.array.reductions` for details of the parameters.

    :Returns:

        `dask.array.Array`
            The unique values.

    """
    x = (
        deepmap(lambda pair: pair["unique"], pairs)
        if not computing_meta
        else pairs
    )
    if computing_meta:
        return x

    x = _concatenate2(x, axes=[0])
    return np.unique(x)


# --------------------------------------------------------------------
# variance
# --------------------------------------------------------------------
def cf_var_chunk(
    x, weights=None, dtype="f8", computing_meta=False, ddof=None, **kwargs
):
    """Chunk calculations for the variance.

    This function is passed to `dask.array.reduction` as its *chunk*
    parameter.

    See
    https://en.wikipedia.org/wiki/Pooled_variance#Sample-based_statistics
    for details.

    .. versionadded:: TODODASKVER

    :Parameters:

        ddof: number
            The delta degrees of freedom. The number of degrees of
            freedom used in the calculation is (N-*ddof*) where N
            represents the number of non-missing elements. A value of
            1 applies Bessel's correction.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dict`
            Dictionary with the keys:

            * N: The sample size.
            * V1: The sum of ``weights`` (equal to ``N`` if weights
                  are not set).
            * V2: The sum of ``weights**2``, or `None` of not
                  required.
            * sum: The weighted sum of ``x``.
            * part: ``V1 * (sigma**2 + mu**2)``, where ``sigma**2`` is
                    the weighted biased (i.e. ``ddof=0``) variance of
                    ``x``, and ``mu`` is the weighted mean of ``x``.
            * weighted: True if weights have been set.
            * ddof: The delta degrees of freedom.

    """
    if computing_meta:
        return x

    weighted = weights is not None

    # N, V1, sum
    d = cf_mean_chunk(x, weights, dtype=dtype, **kwargs)

    wsum = d["sum"]
    V1 = d["V1"]

    avg = divide(wsum, V1, dtype=dtype)
    part = x - avg
    part *= part
    if weighted:
        part = part * weights

    part = chunk.sum(part, dtype=dtype, **kwargs)
    part = part + avg * wsum

    d["part"] = part

    if weighted and ddof == 1:
        d["V2"] = sum_weights_chunk(x, weights, square=True, **kwargs)
    else:
        d["V2"] = None

    d["weighted"] = weighted
    d["ddof"] = ddof

    return d


def cf_var_combine(
    pairs, axis=None, dtype="f8", computing_meta=False, **kwargs
):
    """Combination calculations for the variance.

    This function is passed to `dask.array.reduction` as its *combine*
    parameter.

    .. versionadded:: TODODASKVER

    :Parameters:

        See `dask.array.reductions` for details of the parameters.

    :Returns:

        As for `cf_var_chunk`.

    """
    if not isinstance(pairs, list):
        pairs = [pairs]

    d = next(flatten(pairs))
    weighted = d["weighted"]
    ddof = d["ddof"]
    d = {"weighted": weighted, "ddof": ddof}

    d["part"] = sum_arrays(
        pairs, "part", axis, dtype, computing_meta, **kwargs
    )
    if computing_meta:
        return d["part"]

    d["sum"] = sum_arrays(pairs, "sum", axis, dtype, **kwargs)

    d["N"] = sum_sample_sizes(pairs, axis, **kwargs)
    d["V1"] = d["N"]
    d["V2"] = None
    if weighted:
        d["V1"] = sum_arrays(pairs, "V1", axis, dtype, **kwargs)
        if ddof == 1:
            d["V2"] = sum_arrays(pairs, "V2", axis, dtype, **kwargs)

    return d


def cf_var_agg(
    pairs,
    axis=None,
    dtype="f8",
    computing_meta=False,
    mtol=None,
    original_shape=None,
    **kwargs,
):
    """Aggregation calculations for the variance.

    This function is passed to `dask.array.reduction` as its
    *aggregate* parameter.

    .. note:: Weights are interpreted as reliability weights, as
              opposed to frequency weights.

    See
    https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Reliability_weights
    for details.

    .. versionadded:: TODODASKVER

    :Parameters:

        mtol: number, optional
            The sample size threshold below which collapsed values are
            set to missing data. See `mask_small_sample_size` for
            details.

        original_shape: `tuple`
            The shape of the original, uncollapsed data.

        See `dask.array.reductions` for details of the other
        parameters.

    :Returns:

        `dask.array.Array`
            The collapsed array.

    """
    d = cf_var_combine(pairs, axis, dtype, computing_meta, **kwargs)
    if computing_meta:
        return d

    ddof = d["ddof"]
    V1 = d["V1"]
    wsum = d["sum"]
    var = d["part"] - wsum * wsum / V1

    # Note: var is now the global value of V1 * sigma**2, where sigma
    #       is the global weighted biased (i.e. ddof=0) variance.

    if ddof is None:
        raise ValueError(f"Must set ddof to a numeric value. Got: {ddof!r}")

    if not ddof:
        # Weighted or unweighted variance with ddof=0
        f = 1 / V1
    elif not d["weighted"]:
        # Unweighted variance with any non-zero value of ddof
        f = 1 / (V1 - ddof)
    elif ddof == 1:
        # Weighted variance with ddof=1
        f = V1 / (V1 * V1 - d["V2"])
    else:
        raise ValueError(
            "Can only calculate a weighted variance with ddof=0 or ddof=1. "
            f"Got: {ddof!r}"
        )

    # Now get the required global variance with the requested ddof
    var = f * var

    var = mask_small_sample_size(var, d["N"], axis, mtol, original_shape)
    return var



import cfdm


class CoordinateConversion(cfdm.CoordinateConversion):
    """A coordinate conversion component of a coordinate reference
    construct of the CF data model.

    A coordinate conversion formula converting coordinate values taken
    from the dimension or auxiliary coordinate constructs to a
    different coordinate system. A term of the conversion formula can
    be a scalar or vector parameter which does not depend on any
    domain axis constructs, may have units, or may be a descriptive
    string (such as the projection name "mercator"), or it can be a
    reference to a domain ancillary construct (such as one containing
    spatially varying orography data). A coordinate reference
    construct relates the coordinate values of the field to locations
    in a planetary reference frame.

    .. versionadded:: 3.0.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



import copy
import datetime
import faulthandler
import re
import unittest

import numpy

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class QueryTest(unittest.TestCase):
    f = cf.example_field(1)

    def test_Query_contains(self):

        c = self.f.dim("X")
        self.assertTrue(
            (
                (cf.contains(-4.26) == c).array
                == numpy.array([0, 1, 0, 0, 0, 0, 0, 0, 0], bool)
            ).all()
        )
        self.assertTrue(
            (
                (cf.contains(999) == c).array
                == numpy.array([0, 0, 0, 0, 0, 0, 0, 0, 0], bool)
            ).all()
        )

    def test_Query(self):
        q = cf.Query("wi", [2, 5])
        r = cf.Query("le", 67)
        s = q | r
        t = cf.Query("gt", 12, attr="bounds")
        u = s & t

        repr(q)
        repr(s)
        repr(t)
        repr(u)
        str(q)
        str(s)
        str(t)
        str(u)
        u.dump(display=False)

        u.attr
        u.operator
        q.attr
        q.operator
        q.value
        with self.assertRaises(Exception):
            u.value

        self.assertTrue(u.equals(u.copy(), verbose=2))
        self.assertFalse(u.equals(t, verbose=0))

        copy.deepcopy(u)

        c = self.f.dimension_coordinate("X")
        self.assertTrue(
            (
                (cf.contains(-4.26) == c).array
                == numpy.array([0, 1, 0, 0, 0, 0, 0, 0, 0], bool)
            ).all()
        )
        self.assertTrue(
            (
                (cf.contains(999) == c).array
                == numpy.array([0, 0, 0, 0, 0, 0, 0, 0, 0], bool)
            ).all()
        )

        cf.cellsize(34)
        cf.cellsize(q)

        cf.celllt(3)
        cf.cellle(3)
        cf.cellge(3)
        cf.cellgt(3)
        cf.cellwi(1, 2)
        cf.cellwo(1, 2)

    def test_Query_object_units(self):
        """Check units are processed correctly in and from queries."""

        equivalent_units = {
            1000: ("m", "km"),
            60: ("s", "minute"),
        }  # keys are conversion factors; only use exact equivalents for test
        for conversion, equivalents in equivalent_units.items():
            s_unit, l_unit = equivalents  # s for smaller, l for larger

            # Case 1: only specify units to one component
            q1 = cf.Query("gt", cf.Data(1, s_unit))
            q2 = cf.Query("ge", 1, units=s_unit)
            # Case 2: specify *equal* units across the two components
            q3 = cf.Query("lt", cf.Data(1, s_unit), units=s_unit)
            # Case 3: specify *equivalent* units across the two components
            q4 = cf.Query("le", cf.Data(1, s_unit), units=l_unit)
            # See also final Case 4, below.

            # A) test processed correctly inside unit itself
            for q in [q1, q2, q3, q4]:
                self.assertIsInstance(q, cf.query.Query)
                # TODO: should q4 should return s_ or l_unit? ATM is s_unit.
                self.assertEqual(q._value.Units._units, s_unit)

            with self.assertRaises(ValueError):
                # Case 4: provide non-equivalent units i.e. non-sensical query
                cf.Query("le", cf.Data(1, "m"), units="s")

            # B) Check units are processed correctly when a Query is evaluated
            q5 = cf.Query("eq", conversion, units=s_unit)
            # Pre-comparison, values should be converted for consistent units
            self.assertTrue(
                q5.evaluate(cf.Data([1, 2], l_unit)).equals(
                    cf.Data([True, False])
                )
            )

    def test_Query_as_where_condition(self):
        """Check queries work correctly as conditions in 'where'
        method."""
        # TODO: extend test; added as-is to capture a specific bug (now fixed)

        s_data = cf.Data([30, 60, 90], "second")

        m_lt_query = cf.lt(1, units="minute")
        s_lt_query = cf.lt(60, units="second")
        m_ge_query = cf.ge(1, units="minute")
        s_ge_query = cf.ge(60, units="second")
        for query_pair in [(m_lt_query, s_lt_query), (m_ge_query, s_ge_query)]:
            m_query, s_query = query_pair

            equal_units_where = s_data.data.where(s_query, 0)
            mixed_units_where = s_data.data.where(m_query, 0)
            self.assertTrue(
                (mixed_units_where.array == equal_units_where.array).all()
            )

            equal_units_where_masked = s_data.data.where(s_query, cf.masked)
            mixed_units_where_masked = s_data.data.where(m_query, cf.masked)
            self.assertEqual(
                mixed_units_where_masked.count(),
                equal_units_where_masked.count(),
            )

    def test_Query_datetime1(self):
        d = cf.Data(
            [[1.0, 5.0], [6, 2]],
            "days since 2000-12-29 21:00:00",
            calendar="standard",
        )

        message = "Diff =" + str(
            (
                d - cf.Data(cf.dt("2001-01-03 21:00:00", calendar="standard"))
            ).array
        )

        self.assertTrue(
            (d == cf.eq(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, True], [False, False]]), verbose=2
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ne(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, False], [True, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ge(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, True], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.gt(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, False], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.le(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, True], [False, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.lt(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, False], [False, True]])
            ),
            message,
        )
        self.assertTrue(
            (
                d
                == cf.wi(
                    cf.dt("2000-12-31 21:00:00"), cf.dt("2001-01-03 21:00:00")
                )
            ).equals(cf.Data([[False, True], [False, True]])),
            message,
        )
        self.assertTrue(
            (
                d
                == cf.wo(
                    cf.dt("2000-12-31 21:00:00"), cf.dt("2001-01-03 21:00:00")
                )
            ).equals(cf.Data([[True, False], [True, False]])),
            message,
        )
        self.assertTrue(
            (
                d
                == cf.set(
                    [
                        cf.dt("2000-12-31 21:00:00"),
                        cf.dt("2001-01-03 21:00:00"),
                    ]
                )
            ).equals(cf.Data([[False, True], [False, True]])),
            message,
        )

        cf.seasons()
        [
            cf.seasons(n, start)
            for n in [1, 2, 3, 4, 6, 12]
            for start in range(1, 13)
        ]
        with self.assertRaises(Exception):
            cf.seasons(13)
        with self.assertRaises(Exception):
            cf.seasons(start=8.456)

        cf.mam()
        cf.djf()
        cf.jja()
        cf.son()

    def test_Query_year_month_day_hour_minute_second(self):
        d = cf.Data(
            [[1.0, 5.0], [6, 2]],
            "days since 2000-12-29 21:57:57",
            calendar="gregorian",
        )

        self.assertTrue(
            (d == cf.year(2000)).equals(
                cf.Data([[True, False], [False, True]])
            )
        )
        self.assertTrue(
            (d == cf.month(12)).equals(cf.Data([[True, False], [False, True]]))
        )
        self.assertTrue(
            (d == cf.day(3)).equals(cf.Data([[False, True], [False, False]]))
        )
        d = cf.Data([[1.0, 5], [6, 2]], "hours since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.hour(2)).equals(cf.Data([[False, True], [False, False]]))
        )
        d = cf.Data([[1.0, 5], [6, 2]], "minutes since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.minute(2)).equals(
                cf.Data([[False, True], [False, False]])
            )
        )
        d = cf.Data([[1.0, 5], [6, 2]], "seconds since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.second(2)).equals(
                cf.Data([[False, True], [False, False]])
            )
        )
        d = cf.Data([[1.0, 5.0], [6, 2]], "days since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.year(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )
        self.assertTrue(
            (d == cf.month(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )
        self.assertTrue(
            (d == cf.day(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )
        d = cf.Data([[1.0, 5], [6, 2]], "hours since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.hour(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )
        d = cf.Data([[1.0, 5], [6, 2]], "minutes since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.minute(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )
        d = cf.Data([[1.0, 5], [6, 2]], "seconds since 2000-12-29 21:57:57")
        self.assertTrue(
            (d == cf.second(cf.ne(-1))).equals(
                cf.Data([[True, True], [True, True]])
            )
        )

    def test_Query_dteq_dtne_dtge_dtgt_dtle_dtlt(self):
        d = cf.Data([[1.0, 5.0], [6, 2]], "days since 2000-12-29 21:00:00")

        message = "Diff =" + str(
            (
                d - cf.Data(cf.dt("2001-01-03 21:00:00", calendar="standard"))
            ).array
        )

        self.assertTrue(
            (d == cf.eq(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, True], [False, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ne(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, False], [True, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ge(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, True], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.gt(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[False, False], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.le(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, True], [False, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.lt(cf.dt("2001-01-03 21:00:00"))).equals(
                cf.Data([[True, False], [False, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.eq(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[False, True], [False, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ne(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[True, False], [True, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.ge(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[False, True], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.gt(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[False, False], [True, False]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.le(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[True, True], [False, True]])
            ),
            message,
        )
        self.assertTrue(
            (d == cf.lt(cf.dt(2001, 1, 3, 21, 0, 0))).equals(
                cf.Data([[True, False], [False, True]])
            ),
            message,
        )

        d = cf.dt(2002, 6, 16)
        self.assertEqual(cf.eq(cf.dt(2002, 6, 16)), d)
        self.assertNotEqual(cf.eq(cf.dt(1990, 1, 1)), d)
        self.assertNotEqual(cf.eq(cf.dt("2100-1-1")), d)
        self.assertNotEqual(
            cf.eq(cf.dt("2001-1-1")) & cf.eq(cf.dt(2010, 12, 31)), d
        )

        d = cf.dt(2002, 6, 16)
        self.assertEqual(cf.ge(cf.dt(1990, 1, 1)), d)
        self.assertEqual(cf.ge(cf.dt(2002, 6, 16)), d)
        self.assertNotEqual(cf.ge(cf.dt("2100-1-1")), d)
        self.assertNotEqual(
            cf.ge(cf.dt("2001-1-1")) & cf.ge(cf.dt(2010, 12, 31)), d
        )

        d = cf.dt(2002, 6, 16)
        self.assertEqual(cf.gt(cf.dt(1990, 1, 1)), d)
        self.assertNotEqual(cf.gt(cf.dt(2002, 6, 16)), d)
        self.assertNotEqual(cf.gt(cf.dt("2100-1-1")), d)
        self.assertEqual(
            cf.gt(cf.dt("2001-1-1")) & cf.le(cf.dt(2010, 12, 31)), d
        )

        d = cf.dt(2002, 6, 16)
        self.assertEqual(cf.ne(cf.dt(1990, 1, 1)), d)
        self.assertNotEqual(cf.ne(cf.dt(2002, 6, 16)), d)
        self.assertEqual(cf.ne(cf.dt("2100-1-1")), d)
        self.assertEqual(
            cf.ne(cf.dt("2001-1-1")) & cf.ne(cf.dt(2010, 12, 31)), d
        )

        d = cf.dt(2002, 6, 16)
        self.assertNotEqual(cf.le(cf.dt(1990, 1, 1)), d)
        self.assertEqual(cf.le(cf.dt(2002, 6, 16)), d)
        self.assertEqual(cf.le(cf.dt("2100-1-1")), d)
        self.assertNotEqual(
            cf.le(cf.dt("2001-1-1")) & cf.le(cf.dt(2010, 12, 31)), d
        )

        d = cf.dt(2002, 6, 16)
        self.assertNotEqual(cf.lt(cf.dt(1990, 1, 1)), d)
        self.assertNotEqual(cf.lt(cf.dt(2002, 6, 16)), d)
        self.assertEqual(cf.lt(cf.dt("2100-1-1")), d)
        self.assertNotEqual(
            cf.lt(cf.dt("2001-1-1")) & cf.lt(cf.dt(2010, 12, 31)), d
        )

    def test_Query_evaluate(self):
        for x in (5, cf.Data(5, "kg m-2"), cf.Data([5], "kg m-2 s-1")):
            self.assertEqual(x, cf.eq(5))
            self.assertEqual(x, cf.lt(8))
            self.assertEqual(x, cf.le(8))
            self.assertEqual(x, cf.gt(3))
            self.assertEqual(x, cf.ge(3))
            self.assertEqual(x, cf.wi(3, 8))
            self.assertEqual(x, cf.wo(8, 11))
            self.assertEqual(x, cf.set([3, 5, 8]))

            self.assertEqual(cf.eq(5), x)
            self.assertEqual(cf.lt(8), x)
            self.assertEqual(cf.le(8), x)
            self.assertEqual(cf.gt(3), x)
            self.assertEqual(cf.ge(3), x)
            self.assertEqual(cf.wi(3, 8), x)
            self.assertEqual(cf.wo(8, 11), x)
            self.assertEqual(cf.set([3, 5, 8]), x)

            self.assertNotEqual(x, cf.eq(8))
            self.assertNotEqual(x, cf.lt(3))
            self.assertNotEqual(x, cf.le(3))
            self.assertNotEqual(x, cf.gt(8))
            self.assertNotEqual(x, cf.ge(8))
            self.assertNotEqual(x, cf.wi(8, 11))
            self.assertNotEqual(x, cf.wo(3, 8))
            self.assertNotEqual(x, cf.set([3, 8, 11]))

            # Test that the evaluation is commutative i.e. A == B
            # means B == A
            self.assertNotEqual(x, cf.eq(8))
            self.assertNotEqual(x, cf.lt(3))
            self.assertNotEqual(x, cf.le(3))
            self.assertNotEqual(x, cf.gt(8))
            self.assertNotEqual(x, cf.ge(8))
            self.assertNotEqual(x, cf.wi(8, 11))
            self.assertNotEqual(x, cf.wo(3, 8))
            self.assertNotEqual(x, cf.set([3, 8, 11]))

            self.assertNotEqual(cf.eq(8), x)
            self.assertNotEqual(cf.lt(3), x)
            self.assertNotEqual(cf.le(3), x)
            self.assertNotEqual(cf.gt(8), x)
            self.assertNotEqual(cf.ge(8), x)
            self.assertNotEqual(cf.wi(8, 11), x)
            self.assertNotEqual(cf.wo(3, 8), x)
            self.assertNotEqual(cf.set([3, 8, 11]), x)

        c = cf.wi(2, 4)
        d = cf.wi(6, 8)

        e = d | c

        self.assertTrue(c.evaluate(3))
        self.assertFalse(c.evaluate(5))

        self.assertTrue(e.evaluate(3))
        self.assertTrue(e.evaluate(7))
        self.assertFalse(e.evaluate(5))

        self.assertEqual(3, c)
        self.assertNotEqual(5, c)

        self.assertEqual(c, 3)
        self.assertNotEqual(c, 5)

        self.assertEqual(3, e)
        self.assertEqual(7, e)
        self.assertNotEqual(5, e)

        self.assertEqual(e, 3)
        self.assertEqual(e, 7)
        self.assertNotEqual(e, 5)

        x = "qwerty"
        self.assertEqual(x, cf.eq("qwerty"))
        self.assertEqual(x, cf.eq(re.compile("^qwerty$")))
        self.assertEqual(x, cf.eq(re.compile("qwe")))
        self.assertEqual(x, cf.eq(re.compile("qwe.*")))
        self.assertEqual(x, cf.eq(re.compile(".*qwe")))
        self.assertEqual(x, cf.eq(re.compile(".*rty")))
        self.assertEqual(x, cf.eq(re.compile(".*rty$")))
        self.assertEqual(x, cf.eq(re.compile("^.*rty$")))
        self.assertEqual(x, cf.eq(re.compile("rty$")))

        self.assertNotEqual(x, cf.eq("QWERTY"))
        self.assertNotEqual(x, cf.eq(re.compile("QWERTY")))
        self.assertNotEqual(x, cf.eq(re.compile("^QWERTY$")))
        self.assertNotEqual(x, cf.eq(re.compile("QWE")))
        self.assertNotEqual(x, cf.eq(re.compile("QWE.*")))
        self.assertNotEqual(x, cf.eq(re.compile(".*QWE")))
        self.assertNotEqual(x, cf.eq(re.compile(".*RTY")))
        self.assertNotEqual(x, cf.eq(re.compile(".*RTY$")))
        self.assertNotEqual(x, cf.eq(re.compile("^.*RTY$")))

    def test_Query_set_condition_units(self):
        q = cf.lt(9)
        q.set_condition_units("km")
        self.assertEqual(q.value.Units, cf.Units("km"))

        with self.assertRaises(ValueError):
            q.set_condition_units("seconds")

        q = cf.lt(9000, units="m")
        q.set_condition_units("km")
        self.assertEqual(q.value.Units, cf.Units("km"))
        self.assertEqual(q.value.array, 9)

        q = cf.lt(9)
        r = cf.ge(3000, units="m")
        s = q & r
        s.set_condition_units("km")
        self.assertEqual(s._compound[0].value.Units, cf.Units("km"))
        self.assertEqual(s._compound[1].value.Units, cf.Units("km"))
        self.assertEqual(s._compound[0].value.array, 9)
        self.assertEqual(s._compound[1].value.array, 3)

        self.assertEqual(r.value.Units, cf.Units("m"))
        self.assertEqual(r.value.array, 3000)
        self.assertEqual(q.value, 9)

    def test_Query_iscontains(self):
        q = cf.contains(9)
        self.assertTrue(q.iscontains())

        r = cf.lt(9)
        self.assertFalse(r.iscontains())
        self.assertFalse((q | r).iscontains())

    def test_Query__and__(self):
        q = cf.gt(9)
        r = cf.lt(11)
        s = q & r
        self.assertTrue(s == 10)
        self.assertFalse(s == 8)
        with self.assertRaises(AttributeError):
            s.value

        r = cf.lt(9)
        s = q & r
        self.assertEqual(s.value, 9)
        self.assertFalse(s == 9)

    def test_Query__or__(self):
        q = cf.eq(9)
        r = cf.gt(11)
        s = q | r
        self.assertTrue(s == 9)
        self.assertTrue(s == 12)
        self.assertFalse(s == 10)
        with self.assertRaises(AttributeError):
            s.value

        r = cf.gt(9)
        s = q | r
        self.assertEqual(s.value, 9)
        self.assertFalse(s == 8)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class CellMethodTest(unittest.TestCase):
    strings = (
        "t: mean",
        "time: point",
        "time: maximum",
        "time: sum",
        "lon: maximum time: mean",
        "time: mean lon: maximum",
        "lat: lon: standard_deviation",
        "lon: standard_deviation lat: standard_deviation",
        "time: standard_deviation (interval: 1 day)",
        "area: mean",
        "lon: lat: mean",
        "time: variance (interval: 1 hr comment: sampled " "instantaneously)",
        "time: mean",
        "time: mean time: maximum",
        "time: mean within years time: maximum over years",
        "time: mean within days time: maximum within years time: "
        "variance over years",
        "time: standard_deviation (interval: 1 day)",
        "time: standard_deviation (interval: 1 year)",
        "time: standard_deviation (interval: 30 year)",
        "time: standard_deviation (interval: 1.0 year)",
        "time: standard_deviation (interval: 30.0 year)",
        "lat: lon: standard_deviation (interval: 10 km)",
        "lat: lon: standard_deviation " "(interval: 10 km interval: 10 km)",
        "lat: lon: standard_deviation "
        "(interval: 0.1 degree_N interval: 0.2 degree_E)",
        "lat: lon: standard_deviation "
        "(interval: 0.123 degree_N interval: 0.234 degree_E)",
        "time: variance (interval: 1 hr comment: " "sampled instantaneously)",
        "area: mean where land",
        "area: mean where land_sea",
        "area: mean where sea_ice over sea",
        "area: mean where sea_ice over sea",
        "time: minimum within years time: mean over years",
        "time: sum within years time: mean over years",
        "time: mean within days time: mean over days",
        "time: minimum within days time: sum over days",
        "time: minimum within days time: maximum over days",
        "time: mean within days",
        "time: sum within days time: maximum over days",
    )

    def test_CellMethod__repr__str__(self):
        for s in self.strings:
            cms = cf.CellMethod.create(s)
            t = " ".join(map(str, cms))
            self.assertEqual(t, s, f"{t!r} != {s!r}")
            for cm in cms:
                repr(cm)

    def test_CellMethod_equals(self):
        for s in self.strings:
            cms = cf.CellMethod.create(s)
            for cm in cms:
                self.assertTrue(cm.equals(cm.copy(), verbose=2))

    def test_CellMethod_equivalent(self):
        for s in self.strings:
            cms = cf.CellMethod.create(s)
            for cm in cms:
                self.assertTrue(cm.equivalent(cm.copy(), verbose=2))

        # Intervals
        for s0, s1 in (
            [
                "lat: lon: mean (interval: 10 km)",
                "lat: lon: mean (interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 10 km)",
                "lat: lon: mean (interval: 10 km interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 10 km interval: 10 km)",
                "lat: lon: mean (interval: 10 km interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 20 km interval: 10 km)",
                "lat: lon: mean (interval: 20 km interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 20 km interval: 10 km)",
                "lat: lon: mean (interval: 20000 m interval: 10000 m)",
            ],
            [
                "lat: lon: mean (interval: 10 km)",
                "lon: lat: mean (interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 10 km)",
                "lon: lat: mean (interval: 10 km interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 10 km interval: 10 km)",
                "lon: lat: mean (interval: 10 km interval: 10 km)",
            ],
            [
                "lat: lon: mean (interval: 20 km interval: 10 km)",
                "lon: lat: mean (interval: 10 km interval: 20 km)",
            ],
            [
                "lat: lon: mean (interval: 20 km interval: 10 km)",
                "lon: lat: mean (interval: 10000 m interval: 20000 m)",
            ],
        ):
            cms0 = cf.CellMethod.create(s0)
            cms1 = cf.CellMethod.create(s1)

            for cm0, cm1 in zip(cms0, cms1):
                self.assertTrue(
                    cm0.equivalent(cm1, verbose=2),
                    f"{cm0!r} not equivalent to {cm1!r}",
                )

    def test_CellMethod_get_set_delete(self):
        cm0, cm1 = cf.CellMethod.create(
            "time: minimum within days time: sum over years"
        )

        self.assertEqual(cm0.within, "days")
        self.assertEqual(cm1.over, "years")
        self.assertEqual(cm0.method, "minimum")
        self.assertEqual(cm1.method, "sum")
        self.assertEqual(cm0.axes, ("time",))
        self.assertEqual(cm1.axes, ("time",))

        self.assertIsNone(cm1.get_qualifier("within", None))
        self.assertIsNone(cm0.get_qualifier("where", None))
        self.assertIsNone(cm0.get_qualifier("over", None))

    def test_CellMethod_intervals(self):
        cm = cf.CellMethod.create("lat: mean (interval: 1 hour)")[0]

        self.assertEqual("1 hour", str(cm.intervals[0]))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm

from ..mixin import ArrayMixin


class Array(ArrayMixin, cfdm.Array):
    """Abstract base class for a container of an underlying array.

    The form of the array is defined by the initialisation parameters
    of a subclass.

    .. versionadded:: 3.0.0

    """



from ctypes.util import find_library

from cfunits import Units as cfUnits

_libpath = find_library("udunits2")
if _libpath is None:
    raise FileNotFoundError(
        "cf requires UNIDATA UDUNITS-2. Can't find the 'udunits2' library."
    )


class Units:
    """Store, combine and compare physical units and convert numeric
    values to different units.

    This is a convenience class that creates a `cfunits.Units`
    instance.

    The full documentation is available via a `cf.Units` instance,
    e.g. ``help(cf.Units())``.

    """

    def __new__(cls, *args, **kwargs):
        return cfUnits(*args, **kwargs)

    @staticmethod
    def conform(*args, **kwargs):
        return cfUnits.conform(*args, **kwargs)



import cfdm

from . import mixin
from .decorators import _deprecated_kwarg_check


class CellMeasure(mixin.PropertiesData, cfdm.CellMeasure):
    """A cell measure construct of the CF data model.

    A cell measure construct provides information that is needed about
    the size or shape of the cells and that depends on a subset of the
    domain axis constructs. Cell measure constructs have to be used
    when the size or shape of the cells cannot be deduced from the
    dimension or auxiliary coordinate constructs without special
    knowledge that a generic application cannot be expected to have.

    The cell measure construct consists of a numeric array of the
    metric data which spans a subset of the domain axis constructs,
    and properties to describe the data. The cell measure construct
    specifies a "measure" to indicate which metric of the space it
    supplies, e.g. cell horizontal areas, and must have a units
    property consistent with the measure, e.g. square metres. It is
    assumed that the metric does not depend on axes of the domain
    which are not spanned by the array, along which the values are
    implicitly propagated. CF-netCDF cell measure variables correspond
    to cell measure constructs.

    **NetCDF interface**

    {{netCDF variable}}

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def measure(self):
        """Measure which indicates the metric of space supplied."""
        return self.get_measure(default=AttributeError())

    @measure.setter
    def measure(self, value):
        self.set_measure(value)

    @measure.deleter
    def measure(self):
        self.del_measure(default=AttributeError())

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    @_deprecated_kwarg_check(
        "relaxed_identity", version="3.0.0", removed_at="4.0.0"
    )
    def identity(
        self,
        default="",
        strict=None,
        relaxed=False,
        nc_only=False,
        relaxed_identity=None,
    ):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The `measure` attribute, preceded by ``'measure:'``.
        * The `standard_name` property.
        * The `id` attribute, preceded by ``'id%'``.
        * The `long_name` property, preceded by ``'long_name='``.
        * The netCDF variable name, preceded by ``'ncvar%'``.
        * The value of the *default* parameter.

        .. versionadded:: 3.0.0

        .. seealso:: `id`, `identities`, `long_name`, `measure`,
                     `nc_get_variable`, `standard_name`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

            strict: `bool`, optional
                If True then the identity is the first found of only the
                "measure" attribute, "standard_name" property or the "id"
                attribute.

            relaxed: `bool`, optional
                If True then the identity is the first found of only the
                "measure" attribute, the "standard_name" property, the
                "id" attribute, the "long_name" property or the netCDF
                variable name.

            nc_only: `bool`, optional
                If True then only take the identity from the netCDF
                variable name.

        :Returns:

                The identity.

        **Examples**

        >>> c.measure
        'area'
        >>> c.properties()
        {'long_name': 'cell_area',
         'foo': 'bar'}
        >>> c.nc_get_variable()
        'areacello'
        >>> c.identity()
        'measure:area'
        >>> del c.measure
        >>> c.identity()
        'long_name=cell_area'
        >>> del c.long_name
        >>> c.identity()
        'ncvar%areacello'
        >>> c.nc_del_variable()
        'areacello'
        >>> c.identity()
        ''
        >>> c.identity('no identity')
        'no identity'

        """
        err_msg = "%{0} and 'nc_only' parameters cannot both be True"
        if nc_only:
            if strict:
                raise ValueError(err_msg.format("'strict'"))

            if relaxed:
                raise ValueError(err_msg.format("'relaxed'"))

            n = self.nc_get_variable(None)
            if n is not None:
                return f"ncvar%{n}"

            return default

        n = self.get_measure(default=None)
        if n is not None:
            return f"measure:{n}"

        n = self.get_property("standard_name", None)
        if n is not None:
            return f"{n}"

        n = getattr(self, "id", None)
        if n is not None:
            return f"id%{n}"

        if relaxed:
            n = self.get_property("long_name", None)
            if n is not None:
                return f"long_name={n}"

            n = self.nc_get_variable(None)
            if n is not None:
                return f"ncvar%{n}"

            return default

        if strict:
            return default

        n = self.get_property("long_name", None)
        if n is not None:
            return f"long_name={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default
